import grpc
import uuid
import typing
import asyncio
import logging
import aiocron
import datetime
import dataclasses

from google.protobuf import timestamp_pb2, any_pb2, wrappers_pb2

from perxis.extensions import extension_service_pb2, extension_service_pb2_grpc, extension_pb2
from perxis.roles import roles_pb2_grpc, roles_pb2
from perxis.items import items_pb2_grpc
from perxis.clients import clients_pb2_grpc, clients_pb2
from perxis.common import operation_pb2, operation_service_pb2_grpc, operation_service_pb2, error_pb2
from perxis.collections import collections_pb2_grpc, collections_pb2
from perxis.environments import environments_pb2_grpc
from perxis.extensions.extension_setup import ExtensionSetup


def generate_operation_id() -> str:
    return str(uuid.uuid4())


def datetime_to_timestamp(dt: datetime) -> timestamp_pb2.Timestamp:
    timestamp = dt.timestamp()
    seconds = int(timestamp)
    nanos = int(timestamp % 1 * 1e9)

    return timestamp_pb2.Timestamp(seconds=seconds, nanos=nanos)


@dataclasses.dataclass
class OperationMeta:
    task: asyncio.Task

    operation_id: str
    description: str

    created_at: datetime.datetime = dataclasses.field(default_factory=datetime.datetime.now)
    modified_at: datetime.datetime = dataclasses.field(default_factory=datetime.datetime.now)

    response: typing.Optional[str] = None
    errors: typing.Optional[list[str]] = None

    was_finished: bool = False
    metadata: dict[str, typing.Any] = dataclasses.field(default_factory=dict)

    def mark_cancelled(self):
        self.was_finished = True
        self.response = "Отменено"

    def mark_finished(self, errors: typing.Optional[list[str]] = None):
        self.modified_at = datetime.datetime.now()
        self.was_finished = True
        self.errors = errors
        self.response = None if errors else "OK"

    def to_operation(self):
        packed_response = any_pb2.Any()
        packed_response.Pack(wrappers_pb2.StringValue(value=self.response or "PENDING"))

        return operation_pb2.Operation(
            id=self.operation_id,
            description=self.description,
            created_at=datetime_to_timestamp(self.created_at),
            modified_at=datetime_to_timestamp(self.modified_at),
            done=self.was_finished,
            metadata=self.metadata,
            response=packed_response,
            error=error_pb2.Error(
                message="; ".join(self.errors)
            ) if self.errors else None
        )


class ExtensionService(
    extension_service_pb2_grpc.ExtensionServiceServicer, operation_service_pb2_grpc.OperationServiceServicer
):
    extension_id: str
    collections: list[collections_pb2.Collection] = []
    roles: list[roles_pb2.Role] = []
    clients: list[clients_pb2.Client] = []
    actions: list[dict] = []

    __operations: dict[str, OperationMeta] = {}

    def __init__(self,
                 collections_service: collections_pb2_grpc.CollectionsStub,
                 environments_service: environments_pb2_grpc.EnvironmentsStub,
                 roles_service: roles_pb2_grpc.RolesStub,
                 clients_service: clients_pb2_grpc.ClientsStub,
                 items_service: items_pb2_grpc.ItemsStub,
    ):
        self.logger = logging.getLogger(__name__)
        self.collections_service = collections_service
        self.environments_service = environments_service
        self.roles_service = roles_service
        self.clients_service = clients_service
        self.items_service = items_service

        self.extension_setup = ExtensionSetup(
            self.collections_service, self.environments_service,
            self.roles_service, self.clients_service, self.items_service,
        )

        for collection in self.collections or []:
            self.extension_setup.add_collection(collection)

        for role in self.roles or []:
            self.extension_setup.add_role(role)

        for client in self.clients or []:
            self.extension_setup.add_client(client)

        for action in self.actions or []:
            self.extension_setup.add_action(action)

        @aiocron.crontab('0 * * * *', start=True)
        async def remove_old_operations():
            self.remove_old_operations()

    def remove_old_operations(self):
        self.logger.info("Удаление старых операций")

        ids = []
        now = datetime.datetime.now()

        for operation_meta in self.__operations.values():
            task_is_not_running = operation_meta.task.done() or operation_meta.task.cancelled()

            # Если task фактически не работает то операцию нужно пометить как выполненную. Это может произойти в случае
            # неотловленного исключения, например
            if task_is_not_running:
                operation_meta.was_finished = True

            if not operation_meta.was_finished:
                continue

            time_delta = now - operation_meta.modified_at

            # Для операций которые завершены больше часа назад
            if time_delta.seconds < 60 * 60:
                continue

            # Если по какой-то причине операция была помечена как завершенная но task ещё активен
            need_to_cancel_task = not any(
                (operation_meta.task.done(), operation_meta.task.cancelled())
            )

            if need_to_cancel_task:
                try:
                    operation_meta.task.cancel()
                except Exception as e:
                    self.logger.error("Не удалось отменить task")
                    self.logger.exception(e)

                    continue

            ids.append(operation_meta.operation_id)

            self.logger.info(f"Операция {operation_meta.operation_id} помечена на удаление")

        for operation_id in ids:
            del self.__operations[operation_id]

        self.logger.info("Удаление старых операций завершено")

    def result_log(self, operation, operation_id: str, request, errors: list[str]):
        log_func = self.logger.error if errors else self.logger.info

        log_func(
            "Операция %s (%s) расширения %s для окружения %s пространства %s завершена %s"
            % (
                operation,
                operation_id,
                self.extension_id,
                request.env_id,
                request.space_id,
                "с ошибками: " + "; ".join(errors) if errors else "успешно"
            )
        )

    def ext_request_results_from_exception(self, e: Exception):
        return [
            extension_service_pb2.ExtensionRequestResult(
                extension=self.extension_id,
                state=extension_service_pb2.ExtensionRequestResult.State.ERROR,
                error=str(e),
                msg=None
            )
        ]

    def get_operation_meta(self, operation_id: str) -> typing.Optional[OperationMeta]:
        return self.__operations.get(operation_id)

    def set_operation_meta(self, operation_id: str, description: str, task: asyncio.Task):
        self.__operations[operation_id] = OperationMeta(
            operation_id=operation_id,
            task=task,
            description=description
        )

    async def _Install(self, operation_id: str, request: extension_service_pb2.InstallRequest, context):
        errors_list = await self.extension_setup.install(
            request.space_id, request.env_id, request.force
        )

        self.result_log("установки", operation_id, request, errors_list)

        self.__operations[operation_id].mark_finished(errors_list)

    async def Install(self, request: extension_service_pb2.InstallRequest, context):
        operation_id = generate_operation_id()
        operation_description = "Установка расширения %s для окружения %s пространства %s. %s force" % (
            self.extension_id,
            request.env_id,
            request.space_id,
            "С" if request.force else "Без"
        )

        self.logger.info(operation_description)

        install_task = asyncio.create_task(self._Install(operation_id, request, context))

        self.set_operation_meta(operation_id, operation_description, install_task)

        return self.get_operation_meta(operation_id).to_operation()

    async def _Uninstall(self, operation_id: str, request: extension_service_pb2.UninstallRequest, context):
        errors_list: list[str] = await self.extension_setup.uninstall(request.space_id, request.env_id, request.remove)

        self.result_log("удаления", operation_id, request, errors_list)

        self.__operations[operation_id].mark_finished(errors_list)

    async def Uninstall(self, request: extension_service_pb2.UninstallRequest, context):
        operation_id = generate_operation_id()
        operation_description = "Удаление расширения %s для окружения %s пространства %s. %s remove" % (
            self.extension_id,
            request.env_id,
            request.space_id,
            "С" if request.remove else "Без"
        )

        self.logger.info(operation_description)

        uninstall_task = asyncio.create_task(self._Uninstall(operation_id, request, context))

        self.set_operation_meta(operation_id, operation_description, uninstall_task)

        return self.get_operation_meta(operation_id).to_operation()

    async def _Check(self, operation_id: str, request: extension_service_pb2.CheckRequest, context):
        errors_list: list[str] = await self.extension_setup.check(request.space_id, request.env_id)

        self.result_log("проверки", operation_id, request, errors_list)

        self.__operations[operation_id].mark_finished(errors_list)

    async def Check(self, request: extension_service_pb2.CheckRequest, context):
        operation_id = generate_operation_id()
        operation_description = "Проверка расширения %s для окружения %s пространства %s" % (
            self.extension_id,
            request.env_id,
            request.space_id,
        )

        check_task = asyncio.create_task(self._Check(operation_id, request, context))

        self.set_operation_meta(operation_id, operation_description, check_task)

        return self.get_operation_meta(operation_id).to_operation()

    async def _dispatch_action(
            self, request: extension_pb2.ActionRequest, context
    ) -> extension_service_pb2.ActionResponse:
        action_id = request.action.split("/")[-1]

        func_name = f"action_{action_id}"

        if not hasattr(self, func_name):
            response = extension_service_pb2.ActionResponse(
                state=extension_service_pb2.ActionResponse.State.ERROR,
                title="Невозможно выполнить действие",
                error=f"В расширении отсутсвует функция {action_id}"
            )
        else:
            func = getattr(self, func_name)

            response = await func(request, context)

        return response

    async def Action(self, request: extension_pb2.ActionRequest, context):
        operation_description = "Действие %s для окружения %s пространства %s" % (
            request.action,
            request.env_id,
            request.space_id,
        )

        self.logger.info(operation_description)

        response = await self._dispatch_action(request, context)

        if response.state == extension_service_pb2.ActionResponse.State.ERROR:
            error_description = "Ошибка действия %s в расширении %s: %s" % (
                request.action,
                self.extension_id,
                response.error
            )

            #context.set_code(grpc.StatusCode.UNKNOWN)
            #context.set_details(error_description)

            self.logger.error(error_description)

        return response

    async def Get(self, request: operation_service_pb2.GetOperationRequest, context):
        operations_meta = self.get_operation_meta(request.operation_id)

        if not operations_meta:
            error_description = "Ошибка проверки операции %s в расширении %s - не найдена" % (
                request.operation_id,
                self.extension_id,
            )

            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(error_description)

            self.logger.error(error_description)

            return None

        self.logger.info(
            f"Статус операции {operations_meta.operation_id}:\n"
            f"{operations_meta.description}\n"
            f"{'выполняется' if not operations_meta.was_finished else 'завершена'}, \n"
            f"{'без ошибок' if not operations_meta.errors else 'ошибки: ' + '; '.join(operations_meta.errors)}"
        )

        return operations_meta.to_operation()

    def Cancel(self, request: operation_service_pb2.CancelOperationRequest, context):
        operations_meta = self.get_operation_meta(request.operation_id)

        if not operations_meta:
            error_description = "Не удалось удалить операцию %s в расширении %s - не найдена" % (
                request.operation_id,
                self.extension_id,
            )

            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(error_description)

            self.logger.error(
                error_description
            )

            return None

        try:
            operations_meta.task.cancel()
            operations_meta.mark_cancelled()
        except Exception as e:
            error_description = "Во время отмены операции %s расширении %s произошла ошибка %s" % (
                request.operation_id,
                self.extension_id,
                str(e)
            )

            self.logger.error(error_description)

            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(error_description)

            self.logger.exception(e)

            return None

        self.logger.info(
            "Операция %s в расширении %s отменена" % (
                request.operation_id,
                self.extension_id,
            )
        )

        return operations_meta.to_operation()
